class SenseAPITimeoutException(Exception):
    pass


class SenseAuthenticationException(Exception):
    pass


class SenseMFARequiredException(Exception):
    pass


class SenseWebsocketException(Exception):
    pass


class SenseAPIException(Exception):
    pass
